use std::convert::Infallible;
use std::str::FromStr;

pub(crate) struct Level(pub(crate) Vec<usize>);

impl FromStr for Level {
    type Err = Infallible;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        Ok(Self(
            s.split(" ")
                .filter_map(|value| value.parse::<usize>().ok())
                .collect::<Vec<_>>(),
        ))
    }
}

impl Level {
    // TODO optimize
    pub(crate) fn is_safe(&self) -> bool {
        (0..self.0.len()).all(|index| Self::is_increasing(&self.0[index..]))
            || (0..self.0.len()).all(|index| Self::is_decreasing(&self.0[index..]))
    }
    
    // a b c d e
    // x b c d e
    // a x c d e
    // a b x d e 
    // a b c x e
    // a b c d x
    pub(crate) fn safe_by_removing(&mut self) -> bool {
        if self.is_safe() {
            return true;
        }
        for index in 0..self.0.len() {
            let mut array = self.0.clone();
            array.remove(index);
            if Level(array).is_safe() {
                return true;
            }
        }
        false
    }
    

    pub(crate) fn is_increasing(array: &[usize]) -> bool {
        Self::is_changing(array, PartialOrd::gt)
    }

    pub(crate) fn is_decreasing(array: &[usize]) -> bool {
        Self::is_changing(array, PartialOrd::lt)
    }

    fn is_changing(array: &[usize], operator: fn(&usize, &usize) -> bool) -> bool {
        match array {
            [first, second, ..] if operator(second, first) => Self::is_legal_diff(*first, *second),
            [_singleton] => true,
            _ => false,
        }
    }

    fn is_legal_diff(first: usize, second: usize) -> bool {
        usize::abs_diff(first, second) >= 1 && usize::abs_diff(first, second) <= 3
    }
}
pub(crate) fn parse_input(input: &str) -> Vec<Level> {
    input
        .split("\n")
        .map(Level::from_str)
        .collect::<Result<_, _>>()
        .unwrap()
}

#[cfg(test)]
mod tests {
    use crate::common::Level;

    #[test]
    fn test_is_increasing() {
        let increasing_array = [1, 2, 3];
        assert!(Level::is_increasing(&increasing_array))
    }

    #[test]
    fn test_is_safe_no_change() {
        let increasing_then_not = vec![1, 2, 2, 3];
        assert!(!Level(increasing_then_not).is_safe())
    }

    #[test]
    fn test_task_1() {
        let test_decreasing = vec![7, 6, 4, 2, 1];
        let test_increase_of_5 = vec![1, 2, 7, 8, 9];
        let test_decrease_of_4 = vec![9, 7, 6, 2, 1];
        let test_increase_and_decrease = vec![1, 3, 2, 4, 5];
        let test_leveled = vec![8, 6, 4, 4, 1];
        let test_increasing = vec![1, 3, 6, 7, 9];

        assert!(Level(test_decreasing).is_safe());
        assert!(!Level(test_increase_of_5).is_safe());
        assert!(!Level(test_decrease_of_4).is_safe());
        assert!(!Level(test_increase_and_decrease).is_safe());
        assert!(!Level(test_leveled).is_safe());
        assert!(Level(test_increasing).is_safe());
    }
}
